"""
Error types for muniverse.
"""

class MuniverseError(Exception):
    """
    Exception base class used for muniverse APIs.
    """
    pass

class ProtoError(MuniverseError):
    """
    Low-level protocol error.
    """
    pass

class CallError(MuniverseError):
    """
    Error returned by muniverse API.
    """
    pass



"""
APIs to read and write raw objects in the protocol used by
the muniverse-bind command.
"""

import struct
import bson

from .errors import ProtoError

def read_object(pipe):
    """
    Read and decode a BSON object.
    """
    len_data = pipe.read(4)
    if len(len_data) != 4:
        raise ProtoError('EOF reading length field')
    length = struct.unpack('<I', len_data)[0]
    payload = pipe.read(length)
    if len(payload) != length:
        raise ProtoError('EOF reading BSON object')
    try:
        return bson.loads(payload)
    except ValueError as exc:
        raise ProtoError('BSON decode error: ' + str(exc))

def write_object(pipe, obj):
    """
    Encode and write a BSON object.
    """
    try:
        data = bson.dumps(obj)
    except ValueError as exc:
        raise ProtoError('BSON encode error: ' + str(exc))
    pipe.write(struct.pack('<I', len(data)))
    pipe.write(data)
    pipe.flush()



"""
API for controlling muniverse environments.
"""

import numpy as np

from .handle import Handle

class Env:
    """
    An environment instance.
    """
    def __init__(self, spec, container=None, chrome_host=None, game_host=None):
        """
        Create a new environment from the specification.

        If container is set, launch the environment in a
        custom container.

        If chrome_host and game_host are set, launch the
        environment in a running chrome instance.

        This may raise an exception if a muniverse-bind
        process cannot be started.
        It also raises a MuniverseError if the environment
        cannot be created on the backend.
        """
        if (chrome_host is None) != (game_host is None):
            raise ValueError('must set both chrome_host and game_host')
        elif (not container is None) and (not chrome_host is None):
            raise ValueError('cannot mix chrome_host and container options')
        call_obj = {'Spec': spec}
        call_name = 'NewEnv'
        if not container is None:
            call_obj['Container'] = container
            call_name = 'NewEnvContainer'
        elif not chrome_host is None:
            call_obj['Host'] = chrome_host
            call_obj['GameHost'] = game_host
            call_name = 'NewEnvChrome'
        self.handle = None
        handle = Handle()
        try:
            self.uid = handle.checked_call(call_name, call_obj)['UID']
            self.handle = handle
        finally:
            if self.handle is None:
                handle.close()

    def reset(self):
        """
        Reset the environment to a starting state.
        """
        self.handle.checked_call('Reset', {'UID': self.uid})

    def observe(self):
        """
        Capture a screenshot of the environment.

        The returned value is a 3D numpy array.
        The first index is y, the second is x, the third
        is depth (where depth goes in RGB order).
        """
        res = self.handle.checked_call('Observe', {'UID': self.uid})
        obs = res['Observation']
        dim = [obs['Height'], obs['Width'], 3]
        data = obs['RGB']
        return np.frombuffer(data, dtype=np.uint8).reshape(dim)

    def step(self, seconds, *actions):
        """
        Send actions to the environment and advance time.

        The return value is a tuple (reward, done).
        If done is True, then no more steps can be taken
        until a reset().
        """
        events = []
        for action in actions:
            events.append(action.to_bsonable())
        args = {
            'UID': self.uid,
            'Seconds': seconds,
            'Events': events
        }
        res = self.handle.checked_call('Step', args)
        info = res['StepResult']
        return info['Reward'], info['Done'] != 0

    def close(self):
        """
        Stop and clean up the environment.

        You should not close an environment multiple times.
        """
        try:
            self.handle.checked_call('CloseEnv', {'UID': self.uid})
        finally:
            self.handle.close()



"""
API for dealing with environment specifications.
"""

from .handle import Handle

def spec_for_name(name):
    """
    Find an environment specification for the given name.

    If the specification is not found, None is returned.
    """
    handle = Handle()
    try:
        res = handle.checked_call('SpecForName', {'Name': name})
        if 'Spec' in res:
            return res['Spec']
        return None
    finally:
        handle.close()



"""
Manage and communicate with the muniverse-bind process.
"""

import os
import subprocess
import threading

from . import proto
from .errors import ProtoError, MuniverseError, CallError

# Enable global variables, which are used to share one
# global process between otherwise separate objects.

# pylint: disable=W0603

_CURRENT_SESSION_LOCK = threading.Lock()
_CURRENT_SESSION = None
_CURRENT_SESSION_REFS = 0

class Handle:
    """
    A thread-safe handle on a muniverse-bind instance.
    """
    def __init__(self):
        """
        Create a new handle.

        This may throw a FileNotFoundError or a related
        error if the muniverse-bind command cannot be
        executed.
        """
        global _CURRENT_SESSION
        global _CURRENT_SESSION_LOCK
        global _CURRENT_SESSION_REFS
        _CURRENT_SESSION_LOCK.acquire()
        try:
            if _CURRENT_SESSION is None:
                _CURRENT_SESSION = _Session()
            self.session = _CURRENT_SESSION
            _CURRENT_SESSION_REFS += 1
        finally:
            _CURRENT_SESSION_LOCK.release()
        self.call_lock = threading.Lock()
        self.closed = False

    def close(self):
        """
        Closes the handle.

        This waits for pending calls, and any calls made
        afterwards will fail.

        Calling close more than once has no effect.
        """
        global _CURRENT_SESSION
        global _CURRENT_SESSION_LOCK
        global _CURRENT_SESSION_REFS
        self.call_lock.acquire()
        _CURRENT_SESSION_LOCK.acquire()
        try:
            if self.closed:
                return
            self.closed = True
            _CURRENT_SESSION_REFS -= 1
            if _CURRENT_SESSION_REFS == 0:
                _CURRENT_SESSION.close()
                _CURRENT_SESSION = None
        finally:
            self.call_lock.release()
            _CURRENT_SESSION_LOCK.release()

    def call(self, name, args):
        """
        Make an API call and wait for the result.

        This may throw a MuniverseError or an error
        relating to a closed pipe in the case of a
        communication failure.

        This does not throw an exception for errors
        in the return value itself.
        """
        self.call_lock.acquire()
        try:
            if self.closed:
                raise MuniverseError('call on closed handle')
            return self.session.call(name, args)
        finally:
            self.call_lock.release()

    def checked_call(self, name, args):
        """
        Like call(), but raises a CallError if the call
        results in an error on the back-end.
        """
        res = self.call(name, args)
        if 'Error' in res:
            raise CallError(res['Error'])
        return res

class _Session:
    """
    Manages an instance of muniverse-bind.
    """
    def __init__(self):
        try:
            self.proc = subprocess.Popen(['muniverse-bind'],
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE)
        except OSError:
            gopath = os.environ['GOPATH']
            exc_path = os.path.join(gopath, 'bin', 'muniverse-bind')
            self.proc = subprocess.Popen([exc_path],
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE)
        self.read_thread = threading.Thread(target=self._read_loop)
        self.read_thread.start()
        self.waiting_lock = threading.Lock()
        self.waiting = {}
        self.current_id_lock = threading.Lock()
        self.current_id = 0
        self.send_lock = threading.Lock()

    def close(self):
        """
        Terminate the background process and wait for
        threads to exit.

        This should only be called one time, and it should
        not be called while any other methods are pending.
        """
        self.proc.kill()
        self.read_thread.join()

    def call(self, name, args):
        """
        Make a named API call and wait for the response.
        """
        self.current_id_lock.acquire()
        call_id = str(self.current_id)
        self.current_id += 1
        self.current_id_lock.release()
        call_obj = {'ID': call_id, name: args}
        waiting_obj = {'event': threading.Event()}
        self.waiting_lock.acquire()
        try:
            if self.waiting is None:
                raise MuniverseError('session has died')
            self.waiting[call_id] = waiting_obj
        finally:
            self.waiting_lock.release()
        self.send_lock.acquire()
        try:
            proto.write_object(self.proc.stdin, call_obj)
        finally:
            self.send_lock.release()
        waiting_obj['event'].wait()
        if 'error' in waiting_obj:
            raise waiting_obj['error']
        else:
            return waiting_obj['payload']

    def _read_loop(self):
        try:
            while True:
                payload = proto.read_object(self.proc.stdout)
                if not 'ID' in payload:
                    raise ProtoError('missing ID in response')
                call_id = payload['ID']
                self.waiting_lock.acquire()
                try:
                    if not call_id in self.waiting:
                        raise ProtoError('unregistered call ID: ' + call_id)
                    waiting = self.waiting[call_id]
                    waiting['payload'] = payload
                    waiting['event'].set()
                    del self.waiting[call_id]
                finally:
                    self.waiting_lock.release()
        except MuniverseError as exc:
            self._kill_waiting(exc)

    def _kill_waiting(self, exc):
        self.waiting_lock.acquire()
        try:
            for uid in self.waiting:
                waiting = self.waiting[uid]
                waiting['error'] = exc
                waiting['event'].set()
        finally:
            self.waiting = None
            self.waiting_lock.release()



"""
Types for environment actions.
"""

from .handle import Handle

# Allow more than 5 arguments to a constructor.
# pylint: disable=R0913

def key_for_code(code):
    """
    Create a KeyAction with the given key code.

    The result is None if the code is not found.

    The result will have no 'event', so it will be
    necessary to run with_event() on it.
    """
    handle = Handle()
    try:
        res = handle.checked_call('KeyForCode', {'Code': code})
        if 'KeyEvent' in res:
            action = KeyAction('')
            action.bson_obj = res['KeyEvent']
            return action
        return None
    finally:
        handle.close()

class MouseAction:
    """
    The action type for mouse events.
    """
    def __init__(self, event, x=0, y=0, button="left", click_count=0):
        """
        Create a new mouse event.

        The event is "mousePressed", "mouseReleased",
        or "mouseMoved".
        """
        self.bson_obj = {
            'Type': event,
            'X': x,
            'Y': y,
            'Button': button,
            'ClickCount': click_count
        }

    def to_bsonable(self):
        """
        Create an event object for the back-end.
        """
        return {'MouseEvent': self.bson_obj}

    def with_event(self, event):
        """
        Create a copy of this action with a different
        event name.
        """
        res = MouseAction('', 0, 0)
        res.bson_obj = dict(self.bson_obj, Type=event)
        return res

class KeyAction:
    """
    The action type for keyboard events.
    """
    def __init__(self, event, modifiers=0, timestamp=0, text="", key="",
                 unmodified_text="", key_identifier="", code="",
                 auto_repeat=False, is_keypad=False, is_system_key=False,
                 windows_virtual_key_code=0, native_virtual_key_code=0):
        """
        Create a new keyboard event.

        The event may be "keyDown", "keyUp", or
        "rawKeyDown".
        """
        self.bson_obj = {
            'Type': event,
            'Modifiers': modifiers,
            'Timestamp': timestamp,
            'Text': text,
            'Key': key,
            'UnmodifiedText': unmodified_text,
            'KeyIdentifier': key_identifier,
            'Code': code,
            'AutoRepeat': auto_repeat,
            'IsKeypad': is_keypad,
            'IsSystemKey': is_system_key,
            'WindowsVirtualKeyCode': windows_virtual_key_code,
            'NativeVirtualKeyCode': native_virtual_key_code
        }

    def to_bsonable(self):
        """
        Create an event object for the back-end.
        """
        return {'KeyEvent': self.bson_obj}

    def with_event(self, event):
        """
        Create a copy of this action with a different
        event name.
        """
        res = KeyAction('')
        res.bson_obj = dict(self.bson_obj, Type=event)
        return res
